import argparse

from propargs.type import try_type_val

parser = argparse.ArgumentParser(description='parse key pairs into a dictionary')


def set_props_from_cl(prop_args):
    args, _ = parser.parse_known_args()
    cl_dict = vars(args)['my_dict']
    if not cl_dict:
        return
    for prop_nm in cl_dict :
        arg = cl_dict[prop_nm]
        if prop_nm in prop_args:
            arg = try_type_val(arg, prop_args.props[prop_nm].atype)
        prop_args[prop_nm] = arg


class StoreDictKeyPair(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
         my_dict = {}
         for kv in values.split(","):
             k,v = kv.split("=")
             my_dict[k] = v
         setattr(namespace, self.dest, my_dict)

parser.add_argument("--props", dest="my_dict", action=StoreDictKeyPair, metavar="KEY1=VAL1,KEY2=VAL2...")



from collections import namedtuple

SWITCH = '-'

OS = "OS"

PROPS_DIR = 'props_dir'

UTYPE = "user_type"

# user types (DUMMY type is only for automated test)
TERMINAL = "terminal"
IPYTHON = "iPython"
IPYTHON_NB = "iPython Notebook"
WEB = "Web browser"
DUMMY = "dummy"

VALUE = "val"
QUESTION = "question"
ATYPE = "atype"
HIVAL = "hival"
LOWVAL = "lowval"

BOOL = 'BOOL'
INT = 'INT'
FLT = 'DBL'
STR = 'STR'
CMPLX = 'CMPLX'

AutoLoadProp = namedtuple('AutoLoadProp', 'prop_name, default')
ENV_AUTO_LOAD_PROPS = [AutoLoadProp(prop_name=UTYPE, default=TERMINAL)]



"""
Methods the Data Store import process
"""
import json
import os

from propargs import property_dict
from propargs.constants import PROPS_DIR


def set_props_from_ds(prop_args):
    if prop_args.ds_file:
        ds_dict = _open_file_as_json(prop_args.ds_file)
        property_dict.set_props_from_dict(prop_args, ds_dict)


def _open_file_as_json(ds_file):
    try:
        with open(_path_to_file(ds_file), 'r') as f:
            ds_dict = json.load(f)
    except FileNotFoundError:
        ds_dict = dict()
    return ds_dict


def _path_to_file(ds_file):
    if PROPS_DIR in os.environ:
        path = str(os.environ[PROPS_DIR])
    else:
        path = os.getcwd()

    return os.path.join(path, ds_file)





class Prop:
    """
    Container for prop attributes.

    Attributes include:
        val - the value of the prop
        question - a question prompt for the user's input for the prop value
        atype - the user's answer type (INT, DBL, BOOL, STR, etc.)
        lowval - the lowest value val can take on
        hival - the highest value val can take on
    """

    def __init__(self, val=None, question=None, atype=None,
                 lowval=None, hival=None):
        self.val = val
        self.question = question
        self.atype = atype
        self.lowval = lowval
        self.hival = hival

    def to_json(self):
        return {"val": self.val,
                "question": self.question,
                "atype": self.atype,
                "lowval": self.lowval,
                "hival": self.hival}

    def __str__(self):
        return str(self.val)



"""
prop_args.py
Set, read, and write program-wide properties through this module.
"""
import json
import copy
from typing import Optional

from propargs.exceptions import PropsNotInitializedException
from propargs.prop import Prop
from propargs import data_store, env, property_dict, command_line, user
from propargs.constants import *


def get_prop(key, default_val=None):
    """
    Get a particular property.
    If key is missing (or no props) return default_val.
    """
    if _the_props is None:
        return default_val
    else:
        return _the_props.get(key, default_val)


def set_prop(key, val):
    """
    Set a property's value.
    """
    if _the_props is None:
        raise PropsNotInitializedException("PropArgs not initialized yet.")

    _the_props[key] = val


class PropArgs:
    """
    This class holds named properties for program-wide values.
    It enables getting properties from a file, a database,
    or from the user, either via the command line or a prompt.
    """

    @staticmethod
    def create_props(name, ds_file=None, prop_dict=None,
                     skip_user_questions=False):
        """
        Create a property object with values in 'props'.
        """
        if prop_dict is None:
            prop_dict = {}

        pa = PropArgs(name, ds_file=ds_file, prop_dict=prop_dict,
                      skip_user_questions=skip_user_questions)

        return pa

    def __init__(self, name, logfile=None, ds_file=None,
                 prop_dict=None, skip_user_questions=False):
        """
        Loads and sets properties in the following order:
        1. The Data Store
        2. Property Dictionary
        3. The User's Environment (operating system, dev/prod settings, etc.)
        4. Command Line
        5. Questions Prompts During Run-Time
        """
        global _the_props
        _the_props = self

        self.name = name
        self.logfile = logfile
        self.ds_file = ds_file
        self.props = {}
        if isinstance(prop_dict, dict):
            self.props = copy.deepcopy(prop_dict)

        # 1. The Data Store
        if self.ds_file:
            data_store.set_props_from_ds(self)

        # 2. Property Dictionary
        property_dict.set_props_from_dict(self, prop_dict)

        # 3. The Environment
        env.set_props_from_env(self)

        # 4. process command line args and set them as properties:
        command_line.set_props_from_cl(self)

        if not skip_user_questions and user.can_ask_through_cl(self):
            # 5. Ask the user questions.
            user.ask_user_through_cl(self)

    def get_questions(self):
        all_props = self.to_json()
        question_props = {key: all_props[key] for key in all_props if
                          all_props[key]['question'] is not None}
        return question_props

    def _answer_within_bounds(self, prop_nm, typed_answer):
        if (self.props[prop_nm].atype is None
                or self.props[prop_nm].atype in (STR, BOOL)):
            return True

        if (self.props[prop_nm].lowval is not None
                and self.props[prop_nm].lowval > typed_answer):
            return False

        if (self.props[prop_nm].hival is not None
                and self.props[prop_nm].hival < typed_answer):
            return False

        return True

    def display(self):
        """
        How to represent the properties on screen.
        """
        ret = "Properties in " + self.name + "\n"
        for prop_nm in self:
            ret += "\t" + prop_nm + ": " + str(self.props[prop_nm].val) + "\n"

        return ret

    def __iter__(self):
        return iter(self.props)

    def __str__(self):
        return self.display()

    def __len__(self):
        return len(self.props)

    def __contains__(self, key):
        return key in self.props

    def __setitem__(self, key, v):
        """
        Set a property value.
        """
        if key in self:
            self.props[key].val = v
        else:
            self.props[key] = Prop(val=v)
        if isinstance(v, int):
            self.props[key].atype = INT
        elif isinstance(v, str):
            self.props[key].atype = STR
        elif isinstance(v, bool):
            self.props[key].atype = BOOL
        elif isinstance(v, float):
            self.props[key].atype = FLT

    def __getitem__(self, key):
        return self.props[key].val

    def __delitem__(self, key):
        del self.props[key]

    def items(self):
        return self.props.items()

    def get_logfile(self):
        """
        Special get function for logfile name
        """
        return self.props["log_fname"].val

    def write(self, file_nm):
        """
        Write properties to json file.
        Useful for storing interesting parameter sets.
        """
        dict_for_json = {}
        for prop_name in self.props:
            dict_for_json[prop_name] = self.props[prop_name].to_json()
        f = open(file_nm, 'w')
        json.dump(dict_for_json, f, indent=4)
        f.close()

    def to_json(self):
        return {prop_nm: self.props[prop_nm].to_json() for prop_nm in
                self.props}

    def get(self, key, default=None):
        if key not in self.props or self.props[key].val is None:
            self.props[key] = Prop(val=default)
        return self.props[key].val


_the_props = None



#!/usr/bin/env python3
"""
This is a simple test script. It can be cloned to
create new run scripts, and should be run to test
the system after library changes.
"""
import os
import sys
import pytest
import json

from unittest import mock
from unittest.mock import patch
from propargs import propargs as pa, data_store, env, property_dict, command_line, user
from propargs.constants import *
from propargs.type import try_type_val

DUMMY_PROP_NM = "dummy_prop"
ANSWERS_FOR_INPUT_PROMPTS = [1]


@pytest.fixture
def prop_args():
    """
    A bare-bones propargs object. To use - make `propargs` a test argument.
    """
    return pa.PropArgs.create_props("test_pa", ds_file=None, prop_dict=None, skip_user_questions=True)


@pytest.mark.parametrize('lowval, test_val, hival, expected', [
        (None,  7, None, True),
        (None, -5,   10, True),
        (0,    99, None, True),
        (0,     7,   10, True),
        (0,    77,   10, False),
        (0,    -5,   10, False)
        ])
def test_int_bounds(lowval, test_val, hival, expected, prop_args):
    prop_args.props[DUMMY_PROP_NM] = pa.Prop(lowval=lowval,
                                             hival=hival,
                                             atype=pa.INT)

    assert prop_args._answer_within_bounds(prop_nm=DUMMY_PROP_NM,
                                           typed_answer=test_val) \
           == expected


def test_set_props_from_ds(prop_args):
    with mock.patch('propargs.data_store._open_file_as_json') as mock_open_file_as_json:
        mock_open_file_as_json.return_value = json.loads('{{ "{prop_name}": {{"val": 7, "atype": "INT" }} }}'.format(prop_name=DUMMY_PROP_NM))
        prop_args.ds_file = "some_file"
        data_store.set_props_from_ds(prop_args)
        assert prop_args[DUMMY_PROP_NM] == 7


@pytest.mark.parametrize('environment_variables, ds_file, expected_file_path',
                          [({PROPS_DIR: '/path/to/'}, 'ds_file', '/path/to/ds_file'),
                           (dict(), 'ds_file', os.path.join(os.getcwd(), 'ds_file'))])
def test_path_to_ds_file(environment_variables, ds_file, expected_file_path):

    with mock.patch.dict(os.environ, environment_variables):
        file_path = data_store._path_to_file(ds_file)

    assert file_path == expected_file_path


@mock.patch('propargs.env.ENV_AUTO_LOAD_PROPS', [AutoLoadProp(prop_name="iautoload", default="iamdefault")])
def test_set_props_from_env(prop_args):
    prop_args["prop_also_in_env"] = "imavalue"
    with mock.patch.dict(os.environ,
                         {"prop_also_in_env": "imavalue",
                          "prop_not_yet_loaded": "imanothervalue"}):
        env.set_props_from_env(prop_args)

    assert prop_args["prop_also_in_env"] == "imavalue"
    assert "prop_not_yet_loaded" not in prop_args
    assert prop_args["iautoload"] == "iamdefault"


def test_set_os_in_set_props_from_env(prop_args):
    with mock.patch('platform.system') as mock_platform_system:
        mock_platform_system.return_value = 'Mac'
        env.set_props_from_env(prop_args)

    assert prop_args['OS'] == 'Mac'


def test_props_set_through_prop_dict(prop_args):
    prop_dict = {DUMMY_PROP_NM: {"val": 7, "atype": "INT"} }
    prop_args[DUMMY_PROP_NM] = 100
    property_dict.set_props_from_dict(prop_args, prop_dict)

    assert prop_args[DUMMY_PROP_NM] == 7


def test_type_bool_val():
    assert try_type_val('TRUE', BOOL) is True
    assert try_type_val('yes', BOOL) is True
    assert try_type_val(1, BOOL) is True
    assert try_type_val(1.0, BOOL) is True

    assert try_type_val('false', BOOL) is False
    assert try_type_val('no', BOOL) is False
    assert try_type_val(0, BOOL) is False
    assert try_type_val(0.1, BOOL) is False
    assert try_type_val('abona;a;lnsbaioh', BOOL) is False


def test_type_int_val():
    assert try_type_val('2', INT) == 2


def test_type_flt_val():
    assert try_type_val('2.3', FLT) == 2.3


def test_type_cmplx_val():
    assert try_type_val('3+9j', CMPLX) == 3 + 9j


def test_type_null_val():
    bool_none = try_type_val(None, BOOL)
    float_none = try_type_val(None, FLT)
    int_none = try_type_val(None, INT)
    string_none = try_type_val(None, STR)
    complex_none = try_type_val(None, CMPLX)

    assert isinstance(bool_none, bool)
    assert bool_none is False

    assert isinstance(float_none, float)
    assert float_none == 0.0

    assert isinstance(int_none, int)
    assert int_none == 0

    assert isinstance(string_none, str)
    assert string_none == ''

    assert isinstance(complex_none, complex)
    assert complex_none == 0j


def test_props_set_null_dict(prop_args):
    property_dict.set_props_from_dict(prop_args, None)


def test_prop_set_from_cl(prop_args):
    prop_args.props['existing_prop'] = pa.Prop(atype=pa.INT,
                                               val=-1)

    with patch.object(sys, 'argv', ["file.py", "--irrelevant-switch",
                                    "--props", "existing_prop=7,new_prop=4",
                                    "--other-irrelevant-switch"]):
        command_line.set_props_from_cl(prop_args)

    assert prop_args['existing_prop'] == 7
    assert prop_args['new_prop'] == '4'


def test_user_input(prop_args):
    prop_args.props[DUMMY_PROP_NM] = pa.Prop(atype=pa.INT,
                                             question="Enter Integer: ",
                                             val=-1)

    with patch('builtins.input', side_effect=ANSWERS_FOR_INPUT_PROMPTS):
        user.ask_user_through_cl(prop_args)

    assert prop_args[DUMMY_PROP_NM] == ANSWERS_FOR_INPUT_PROMPTS[0]


def test_get_questions(prop_args):
    prop_args.props['question_prop'] = pa.Prop(question="Enter Integer: ")

    prop_args.props['no_question_prop'] = pa.Prop()

    qs = prop_args.get_questions()

    assert 'question_prop' in qs
    assert 'no_question_prop' not in qs


def test_singleton():
    pa.PropArgs('props1', prop_dict={'i like': {'val': 'turtles'}})

    assert pa.get_prop('i like') == 'turtles'

    pa.PropArgs('props2', prop_dict={'i like': {'val': 'tests that pass'}})

    assert pa.get_prop('i like') == 'tests that pass'


def test_set_prop():
    pa.PropArgs('props1', prop_dict={'i like': {'val': 'turtles'}})

    assert pa.get_prop('i like') == 'turtles'

    pa.set_prop('i like', 'tests that pass')

    assert pa.get_prop('i like') == 'tests that pass'




import os
import platform

from propargs.constants import OS, ENV_AUTO_LOAD_PROPS


def set_props_from_env(prop_args):
    _set_auto_load_props(prop_args)
    _overwrite_existing_props(prop_args)

def _set_auto_load_props(prop_args):
    for prop_name, default in ENV_AUTO_LOAD_PROPS:
        prop_args[prop_name] = os.environ.get(prop_name, default=default)

def _overwrite_existing_props(prop_args):
    env_dict = os.environ
    common_prop_keys = env_dict.keys() & prop_args.props.keys()
    for key in common_prop_keys:
        prop_args[key] = env_dict[key]

    prop_args[OS] = platform.system()




from propargs.constants import *
from propargs.type import try_type_val

def ask_user_through_cl(prop_args):
    for prop_nm in prop_args:
        if (hasattr(prop_args.props[prop_nm], QUESTION)
            and prop_args.props[prop_nm].question):
            prop_args.props[prop_nm].val = _ask_until_correct(prop_args, prop_nm)

def _ask_until_correct(prop_args, prop_nm):
    atype = None
    if hasattr(prop_args.props[prop_nm], ATYPE):
        atype = prop_args.props[prop_nm].atype

    while True:
        answer = input(_get_question(prop_args, prop_nm))
        if not answer:
            return prop_args.props[prop_nm].val

        try:
            typed_answer = try_type_val(answer, atype)
        except ValueError:
            print("Input of invalid type. Should be {atype}"
                  .format(atype=atype))
            continue

        if not prop_args._answer_within_bounds(prop_nm, typed_answer):
            print("Input must be between {lowval} and {hival} inclusive."
                  .format(lowval=prop_args.props[prop_nm].lowval,
                          hival=prop_args.props[prop_nm].hival))
            continue

        return typed_answer

def _get_question(prop_args, prop_nm):
        return "{question} [{lowval}-{hival}] ({default}) "\
               .format(question=prop_args.props[prop_nm].question,
                       lowval=prop_args.props[prop_nm].lowval,
                       hival=prop_args.props[prop_nm].hival,
                       default=prop_args.props[prop_nm].val)

def can_ask_through_cl(prop_args):
    return UTYPE in prop_args and prop_args[UTYPE] in (TERMINAL, IPYTHON, IPYTHON_NB)



from propargs.constants import *
from propargs.type import try_type_val
from propargs.prop import Prop


def set_props_from_dict(prop_args, prop_dict):
    """
    Dict Example:

    {
        "prop_name_1": {
            "val": 1,
            "question": "What value should this property have?",
            "atype": "int",
            "hival": 10,
            "lowval": 0
        },
        "prop_name_2": {
            "val": "Hello World."
        }
    }
    """
    if prop_dict is None:
        return

    for prop_nm in prop_dict:
        atype = prop_dict[prop_nm].get(ATYPE, STR)
        val = try_type_val(prop_dict[prop_nm].get(VALUE, None),
                                      atype)
        question = prop_dict[prop_nm].get(QUESTION, None)
        hival = prop_dict[prop_nm].get(HIVAL, None)
        lowval = prop_dict[prop_nm].get(LOWVAL, None)
        prop_args.props[prop_nm] = Prop(val=val, question=question, atype=atype,
                                   hival=hival, lowval=lowval)



from propargs.constants import *


def try_type_val(val, atype):
    if atype in TYPE_DICT:
        if val is None:
            return TYPE_FOR_NONE[atype]
        else:
            type_cast = TYPE_DICT[atype]
            return type_cast(val)
    else:
        return val


def boolean(val):
    if type(val) is str:
        return val.lower() == 'true' or val.lower() == 'yes'

    if type(val) in (int, float):
        return val == 1

    return bool(val)


TYPE_DICT = {BOOL: boolean, INT: int, FLT: float, CMPLX: complex, STR: str}
TYPE_FOR_NONE = {BOOL: False, INT: 0, FLT: 0.0, CMPLX: 0j, STR: ''}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="propargs",
    version="0.0.20",
    author="Nathan Conroy, Gene Callahan",
    author_email="nathanconroydev@gmail.com",
    description="A module for systematically organizing user preferences "
                "acquired from a database, env vars, a parameter file, "
                "or user choices.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gcallah/PropArgs",
    packages=['propargs'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)
